package propertyGridPack;

import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.lang.reflect.Field;
import java.util.Enumeration;
import java.util.Properties;
import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;

/**
 * The Class PropertyGrid.
 *
 * @author Hroudtwolf
 * @since 03/06/2011
 * @license Creative Commons Attribution 3.0 Unported License.
 */

public class PropertyGrid extends JScrollPane implements ActionListener, MouseListener, FocusListener, ComponentListener
{
	
		
	/**
	 * The Class PropAssoc
	 */
	private class PropAssoc
	{
		
		/** The associated property object. */
		public Properties	objObjectAssoc;
		
		/** The property key. */
		public String		sKey;
		
		/**
		 * Instantiates a new prop assoc.
		 *
		 * @param objObjectAssoc the associated property object
		 * @param sKey the key of the property field.
		 */
		public PropAssoc (Object objObjectAssoc, String sKey)
		{
			
			this.objObjectAssoc		= (Properties) objObjectAssoc;
			this.sKey				= sKey;
			
			return;
		}
	}
	
	/** The Constant LINEHEIGHT */
	private static final int LINEHEIGHT = 20;
	
	/** The Constant LINEMARGIN */
	private static final int LINEMARGIN = 2;
	
	/** The Constant serialVersionUID. */
	private static final long serialVersionUID = -954726092768114184L;
	
	/** The inner container. */
	private JPanel 	objInnerContainer;
	
	/** The incremental inner height. */
	private int 	nInnerHeight;
	
	/** The background color of the labels. */
	private Color	objLabelBackColor;
	
	/** The cursor object of the labels. */
	private Cursor	objLabelCursor	;
	
	/** The property listener. */
	private IPropertyListener objChangeListener;
	
	
	/**
	 * Retrieves the natural sentence of an input string.
	 *
	 * @param sInput the input string
	 * @return a string
	 */
	private String getNaturalSentence (String sInput)
	{
		char 	cChar[] 			= sInput.toCharArray ();
		String 	sOutput 			= "";
		boolean blSenStart			= true;
		boolean blLastWasUnderline 	= false;
		
		for (char cOneChar : cChar)
		{	 
			switch (cOneChar)
			{
				case '_':
					if (blLastWasUnderline)
					{
						blSenStart = true;							;
					}else{
						sOutput += " ";
					}
					blLastWasUnderline = true;
					break;
					
				default:					
					sOutput 			+= blSenStart ? String.valueOf (cOneChar).toUpperCase () : cOneChar;						
					blSenStart 			= false;
					blLastWasUnderline 	= false;
			}		 
		}
		
		return sOutput;
	}

	/**
	 * Adds a category to the current grid.
	 *
	 * @param sTitle the title
	 */
	private void addCategory (String sTitle)
	{
		JLabel objCategory = new JLabel (sTitle);
		
		objCategory.setBounds (0, this.nInnerHeight, this.objInnerContainer.getWidth (), 20);
		objCategory.setBorder (BorderFactory.createEtchedBorder ());
		objCategory.setName ("category");
		this.objInnerContainer.add (objCategory);
		
		this.nInnerHeight = this.nInnerHeight + LINEHEIGHT + LINEMARGIN;
		this.objInnerContainer.setPreferredSize (new Dimension (this.objInnerContainer.getWidth (), this.nInnerHeight));
		
		return;
	}
	
	/**
	 * Adds a text field to the current.
	 *
	 * @param sLabel the label
	 * @param objObjectAssoc the associated property object.
	 * @param sKey the key
	 */
	private void addTextField (String sLabel, Object objObjectAssoc, String sKey) 
	{
		JLabel 		objLabel = new JLabel (sLabel);
		JTextField 	objField = new JTextField (((Properties)objObjectAssoc).getProperty (sKey));
		
		objLabel.setName ("label");
		objField.setName ("field");
		
		objLabel.setOpaque(true);
		objLabel.setBackground	(this.objLabelBackColor);
		objField.setBorder 		(null);
		
		objLabel.putClientProperty ("assoc", objField);
		objField.putClientProperty ("assoc", new PropAssoc (objObjectAssoc, sKey));
		
		objLabel.setBounds (0, this.nInnerHeight, (this.objInnerContainer.getWidth () / 2) - LINEMARGIN, 20);
		objField.setBounds ((this.objInnerContainer.getWidth () / 2), this.nInnerHeight, this.objInnerContainer.getWidth () / 2, 20);

		this.objInnerContainer.add (objLabel);
		this.objInnerContainer.add (objField);
		
		this.nInnerHeight = this.nInnerHeight + LINEHEIGHT + LINEMARGIN;
		this.objInnerContainer.setPreferredSize (new Dimension (this.objInnerContainer.getWidth (), this.nInnerHeight));
		
		objField.addFocusListener   (this);
		objField.addActionListener 	(this);
		objLabel.addMouseListener	(this);
		
		objLabel.setCursor (this.objLabelCursor);
		
		return;
	}
	
	/**
	 * Instantiates a new property grid.
	 *
	 * @param nX the x coordinate
	 * @param nY the y coordinate
	 * @param nWidth the width
	 * @param nHeight the height
	 */
	public PropertyGrid (int nX, int nY, int nWidth, int nHeight)
	{
	    super (JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
	    
	    this.setBounds (nX, nY, nWidth, nHeight);
	    this.setBackground (new Color (215, 215, 215));

	    this.objInnerContainer = new JPanel ();
	    this.objInnerContainer.setLayout (null);
	    this.objInnerContainer.setBounds (0, 0, this.getViewportBorderBounds ().width - (this.getVerticalScrollBar().getPreferredSize().width + this.getViewportBorderBounds ().x), 0);
 
	    this.setViewportView (this.objInnerContainer);
	    this.objInnerContainer.setVisible (true);
 
	    // Stock objects
	    this.objLabelBackColor 	= Color.WHITE;
	    this.objLabelCursor		= new Cursor (Cursor.TEXT_CURSOR);
	    
	    this.addComponentListener (this);
	 
	    
		return;
	}
	
	/**
	 * Sets a listener object which retrieve notification if a property was changed.
	 * The listener class has one method.
	 * If this method returns false, change will not confirmed.
	 * You can use it for developing input filters/validators.
	 */
	public void setPropertyListener (IPropertyListener objChangeListener)
	{
		
		this.objChangeListener = objChangeListener; 
		
		return;
	}
	
	/**
	 * Updates the UI with data from the selected object.
	 * Attention!!! All unconfirmed inputs will reseted.
	 */
	public void updateObject ()
	{
		Component[] objComponents;
		PropAssoc	objAssoc;
		
		objComponents = this.objInnerContainer.getComponents ();
		
		for (Component objSingleCom : objComponents)
		{
			if (objSingleCom.getName ().equals ("field"))
			{
				objAssoc = (PropAssoc) ((JTextField)objSingleCom).getClientProperty ("assoc");
				((JTextField)objSingleCom).setText (objAssoc.objObjectAssoc.getProperty (objAssoc.sKey));
			}		
		}
	
		return;
	}	
	
	/**
	 * Selects an object.
	 * Every field of the "Properties" class type will determined.
	 *
	 * @param objSource the source object
	 */
	public void selectObject (Object objSource)
	{
		@SuppressWarnings("rawtypes")
		Class 		objClass 	= objSource.getClass ();
		Field [] 	objFields 	= objClass.getFields ();
		Properties	objProps	= null;
		@SuppressWarnings("rawtypes")
		Enumeration objEnum		= null;
		String 		sKey;
		
		this.objInnerContainer.removeAll ();
		
		for (Field objSingleField : objFields)
		{
			if (objSingleField.getType ().getCanonicalName ().equals ("java.util.Properties"))
			{
				try {
					objProps 	= (Properties) objSingleField.get (objSource);
					objEnum		= objProps.propertyNames ();
					
					this.addCategory (this.getNaturalSentence (objSingleField.getName ()));
					
				    while (objEnum.hasMoreElements ()) 
				    {
				        sKey = (String) objEnum.nextElement ();

				        this.addTextField (sKey, objProps, sKey); 
				      }
					
				} catch (Exception e) {
					e.printStackTrace ();
				}
				
			}
		}
		
		return;
	}

	/* (non-Javadoc)
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	@Override
	public void actionPerformed (ActionEvent objEvent) 
	{
		JTextField objField = (JTextField) objEvent.getSource ();
		PropAssoc  objAssoc = (PropAssoc) objField.getClientProperty ("assoc");
		
		if (this.objChangeListener != null)
		{
			if (this.objChangeListener.propertyChanged(objAssoc.sKey, objField.getText ()))
			{
				objAssoc.objObjectAssoc.setProperty (objAssoc.sKey, objField.getText ());
			}
		}
		
		return;
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseClicked (MouseEvent objEvent) 
	{
		JTextField objField = (JTextField)((JLabel) objEvent.getSource ()).getClientProperty ("assoc");
		
		objField.requestFocus ();
		
		return;	
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseEntered (MouseEvent objEvent) 
	{

		// Dummy, dummy, dummy... I got love in my tummy.
		
		return;
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseExited (MouseEvent objEvent)
	{
		
		// Dummy, dummy, dummy... I got love in my tummy.

		return;
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	@Override
	public void mousePressed (MouseEvent objEvent)
	{
		
		// Dummy, dummy, dummy... I got love in my tummy.

		return;
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	@Override
	public void mouseReleased (MouseEvent objEvent)
	{
		
		// Dummy, dummy, dummy... I got love in my tummy.
	
		return;	
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	@Override
	public void focusGained (FocusEvent objEvent)
	{

		// Dummy, dummy, dummy... I got love in my tummy.
		
		return;
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	@Override
	public void focusLost (FocusEvent objEvent)
	{
		JTextField objField = (JTextField) objEvent.getSource ();
		PropAssoc  objAssoc = (PropAssoc) objField.getClientProperty ("assoc");
		
		if (this.objChangeListener != null)
		{
			if (this.objChangeListener.propertyChanged(objAssoc.sKey, objField.getText ()))
			{
				objAssoc.objObjectAssoc.setProperty (objAssoc.sKey, objField.getText ());
			}
		}
		
		return;
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	@Override
	public void componentHidden(ComponentEvent objEvent) 
	{
		
		// Dummy, dummy, dummy... I got love in my tummy.
		
		return;
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	@Override
	public void componentMoved(ComponentEvent objEvent) 
	{
		
		// Dummy, dummy, dummy... I got love in my tummy.
		
		return;
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	@Override
	public void componentResized(ComponentEvent objEvent) 
	{
		Component[] objComponents;
			
		this.objInnerContainer.setBounds (0, 0, this.getViewportBorderBounds ().width - this.getViewportBorderBounds ().x, 0);
		
		objComponents = this.objInnerContainer.getComponents ();
		
		for (Component objSingleCom : objComponents)
		{
			if (objSingleCom.getName ().equals ("category"))
			{
				objSingleCom.setSize (this.objInnerContainer.getWidth (), 20);
			}else if (objSingleCom.getName ().equals ("label"))
			{
				objSingleCom.setSize ((this.objInnerContainer.getWidth () / 2), 20);
			}else{
				objSingleCom.setBounds ((this.objInnerContainer.getWidth () / 2), objSingleCom.getY (), this.objInnerContainer.getWidth () / 2, 20);
			}
		}
		
		return;
	}

	/* (non-Javadoc)
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	@Override
	public void componentShown(ComponentEvent objEvent) 
	{
		
		// Dummy, dummy, dummy... I got love in my tummy.
		
		return;
	}
}

/*
 * Object conditions:
 * 
 * You can select each type of object with the property grid.
 * The select method of the property grid will determine every occurrence of 
 * public fields which are typed by the "Properties" class.
 * The name of the respective fields will used for category names.
 * You can add spaces to the names by using the underscore.
 * If you use two underscore behind each other, the next alphabetic letter will
 * set to upper case.
 * The first letter of a field will automatically set to upper case.
 * 
 */
